/* sysvipc.c -- SYSV IPC
   Copyright (c) 1998 by Holger Veit

This file is part of emx.

emx is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

emx is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with emx; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.

As special exception, emx.dll can be distributed without source code
unless it has been changed.  If you modify emx.dll, this exception
no longer applies and you must remove this paragraph from all source
files for emx.dll.  */

#define INCL_BASE
#define INCL_DOSPROCESS
#define INCL_DOSMEMMGR
#define INCL_DOSSEMAPHORES
#include <os2.h>

#include <sys/types.h>
#include <sys/errno.h>
#include <sys/timeb.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <emx/syscalls.h>
#include "emxdll.h"
#include "sysvipc.h"

/**************************************************************************
 * S H M section
 *************************************************************************/
char shmflg = 0;
static HMTX shm_mutex = 0;
static struct _shmcommon *shm_region = 0;

int shm_attach()
{
  int i;
  struct _shmitem *shm;
  int rc;

  if (shmflg)
    return shm_region ? -1 : 0;

  rc = sysv_attachglobal("\\SHAREDMEM\\EMXSHM.MEM",
			  "\\SEM32\\EMXSHM.SEM",
			  sizeof(struct _shmitem)*SHMMNI + sizeof(short)*2,
			  (PPVOID)&shm_region,
			  &shm_mutex);
  shmflg = 1;
  switch (rc) {
  case -1: /* error */
    return -1;
  case 0:  /* opened */
    return 0;
  case 1:  /* created, must be initialized */
    shm_region->seqno = 0;
    shm_region->flags = 0;
    shm_region->usecnt = 0;
    shm_region->maxid = 0;
    shm = shm_region->shm;
    for (i=0; i<SHMMNI; i++) {
      shm[i].flags = 0; /* mark unused */
      shm[i].ds.shm_perm.key = SHM_UNUSED;
    }
  default:
    return 0;
  }
}

void shm_detach()
{
  sysv_detachglobal(shm_region,shm_mutex);
}

int shm_findkey(key_t key)
{
  /* find a slot */
  int i;
  if (key==IPC_PRIVATE) key = SHM_UNUSED; /* IPC_PRIVATE will always create */

  for (i=0; i<SHMMNI; i++) {
    if (shm_region->shm[i].ds.shm_perm.key == key &&
	shm_region->shm[i].flags == 0) {
      return i;
    }
  }
  return -1;
}

#define shm_enter() DosRequestMutexSem(shm_mutex,-1l)
#define shm_leave() DosReleaseMutexSem(shm_mutex)

int shm_error(int *errnop, int val) 
{
  *errnop = val;
  shm_leave();
  return -1;
}

int sysv_shmget(struct _shmargs* args, int *errnop)
{
  /* unpack args */
  key_t key = args->u.get.key;
  int sz = args->u.get.sz;
  int flg = args->u.get.flg;  
  int id,rc;
  struct _shmitem* shm = shm_region->shm;
  struct timeb dst;
  struct shmid_ds* p = 0;

  *errnop = 0;

  if (shm_attach() == -1) {
    *errnop = ENOSYS;
    return -1;
  }

  shm_enter();
  id = shm_findkey(key);
  rc = ipc_open(key, id, flg,
		shm_region->usecnt>=SHMMNI,
		id==-1 ? 0 : &shm[id].ds.shm_perm,
		errnop);
  if (rc==-1) {
    shm_leave();
    return -1;
  }

  /* id is the slot we will use */
  if (id==-1) { /* need to create a new slot */
    id = shm_findkey(SHM_UNUSED);
    if (id==-1)
      return shm_error(errnop,ENOSPC);
  }
  /* now we have a slot id */
  if ((flg & IPC_CREAT) || key==IPC_PRIVATE) {
    /* fill in the relevant data */
    if (id > shm_region->maxid)
      shm_region->maxid = id;
    p = &shm[id].ds;
    p->shm_perm.mode = flg;
    p->shm_perm.cuid = p->shm_perm.uid = ipc_uid();
    p->shm_perm.cgid = p->shm_perm.gid = ipc_gid();
    p->shm_perm.seq = shm_region->seqno;
    p->shm_segsz = sz;
    p->shm_cpid = p->shm_lpid = ipc_pid();
    p->shm_nattch = 0;
    p->shm_atime = p->shm_dtime = 0;
    do_ftime(&dst);
    p->shm_ctime = dst.time;

    rc = DosAllocSharedMem(&shm[id].ptr,NULL,sz,
	OBJ_GETTABLE|PAG_COMMIT|PAG_READ|PAG_WRITE);
    if (rc) 
      return shm_error(errnop,ENOMEM);

    /* make the rest of the table valid */
    p->shm_perm.key = key;
    shm[id].flags |= SHM_INUSE;

    shm_region->usecnt++;
  }

  shm_leave();
  return p->shm_perm.seq * SHMMNI + id;
}

int sysv_shmat(struct _shmargs* args, int *errnop)
{
  /* unpack args */
  int id = args->u.at.id;
  void* addr = args->u.at.addr;
  int flg = args->u.at.flg;

  struct _shmitem *shm;
  struct shmid_ds *p;
  struct timeb dst;
  int rc;

  *errnop = 0;

  /* not loaded yet */
  if (!shmflg) {
    *errnop = ENOSYS;
    return -1;
  }

  shm_enter();

  /* addr specified? OS/2 can't map to a specific address */
  if (addr)
    return shm_error(errnop,EINVAL);

  /* invalid id? */
  if (id < 0)
    return shm_error(errnop,EINVAL);

  id %= SHMMNI;
  shm = &shm_region->shm[id];
  p = &shm->ds;
  
  /* not in use? */
  if ((shm->flags & SHM_INUSE)==0)
    return shm_error(errnop,EINVAL);

  if (ipc_access(&p->shm_perm,
                 flg & SHM_RDONLY ? S_IRUGO : S_IRUGO|S_IWUGO)) {
    return shm_error(errnop,EACCES);
  }
  if (p->shm_perm.seq != (id / SHMMNI))
    return shm_error(errnop,EINVAL);

  p->shm_nattch++;
  p->shm_lpid = ipc_pid();
  do_ftime(&dst);
  p->shm_atime = dst.time;

  rc = DosGetSharedMem(shm[id].ptr,PAG_READ|PAG_WRITE);
  if (rc)
    return shm_error(errnop,ENOMEM);

  shm_leave();
  return (int)shm[id].ptr;
}

int sysv_shmctl(struct _shmargs* args, int *errnop)
{
  /* unpack args */
  int id = args->u.ctl.id;
  int cmd = args->u.ctl.cmd;
  void* buf = args->u.ctl.buf;
  struct shmid_ds* bds = (struct shmid_ds*)buf;
  struct shmid_ds* p = 0;
  struct _shmitem* shm;
  int ids;

  *errnop = 0;

  /* not loaded yet */
  if (!shmflg) {
    *errnop = ENOSYS;
    return -1;
  }

  shm_enter();
  if (id < 0 && cmd < 0)
    return shm_error(errnop,EINVAL);

  switch(cmd) {

  case IPC_INFO: {
    struct shminfo* info = (struct shminfo*)buf;
    if (!info)
      return shm_error(errnop,EFAULT);
    info->shmmni = SHMMNI;
    info->shmmax = SHMMAX;
    info->shmmin = SHMMIN;
    info->shmall = SHMALL;
    info->shmseg = SHMALL;
    shm_leave();
    return shm_region->maxid;
   }

  case SHM_STAT:
    if (!bds)
      return shm_error(errnop,EFAULT);
    if (id > shm_region->maxid)
      return shm_error(errnop,EINVAL);
    shm = &shm_region->shm[id];
    p = &shm->ds;
    if ((shm->flags & SHM_INUSE)==0)
      return shm_error(errnop,EINVAL);
    bds->shm_perm   = p->shm_perm;
    bds->shm_segsz  = p->shm_segsz;
    bds->shm_atime  = p->shm_atime;
    bds->shm_dtime  = p->shm_dtime;
    bds->shm_ctime  = p->shm_ctime;
    bds->shm_cpid   = p->shm_cpid;
    bds->shm_lpid   = p->shm_lpid;
    bds->shm_nattch = p->shm_nattch;
    shm_leave();
    return p->shm_perm.seq*SHMMNI + id;
  }

  if (!bds)
    return shm_error(errnop,EFAULT);

  ids = id;
  shm = &shm_region->shm[ids %= SHMMNI];
  p = &shm->ds;

  if (p->shm_perm.seq != (id/SHMMNI))
    return shm_error(errnop,EINVAL);

  if ((shm->flags & SHM_INUSE)==0)
    return shm_error(errnop,EINVAL);

  switch (cmd) {

  case IPC_RMID:
    shm->flags |= SHM_DELSEG;
    if (p->shm_nattch == 0)
      shm_delseg(shm);
    shm_leave();
    return 0;

  case IPC_STAT:
    if (ipc_access(&p->shm_perm,S_IRUGO))
      return shm_error(errnop,EACCES);
    bds->shm_perm   = p->shm_perm;
    bds->shm_segsz  = p->shm_segsz;
    bds->shm_atime  = p->shm_atime;
    bds->shm_dtime  = p->shm_dtime;
    bds->shm_ctime  = p->shm_ctime;
    bds->shm_cpid   = p->shm_cpid;
    bds->shm_lpid   = p->shm_lpid;
    bds->shm_nattch = p->shm_nattch;
    shm_leave();
    return p->shm_perm.seq*SHMMNI + ids;

  case IPC_SET:
    /* TBD when uid/gid is implemented */
    shm_leave();
    return 0;

  default:
  case SHM_INFO:
  case SHM_LOCK:
  case SHM_UNLOCK:
    return shm_error(errnop,EINVAL);
  }
}

void shm_delseg(struct _shmitem* shm)
{
}

int sysv_shmdt(struct _shmargs* args, int *errnop)
{
  int id;
  struct _shmitem* shm = 0;
  struct shmid_ds* p = 0;
  struct timeb dst;

  /* unpack args */
  void* addr = (void*)args;

  /* not loaded yet */
  if (!shmflg) {
    *errnop = ENOSYS;
    return -1;
  }

  shm_enter();
  for (id=0; id<SHMMNI; id++) {
    void* ptr = shm_region->shm[id].ptr;
    if (ptr>= addr && addr < (ptr+shm_region->shm[id].ds.shm_segsz))
      break;
  }
  if (id==SHMMNI) {
    *errnop = EINVAL;
    shm_leave();
    return -1;
  }
  /* id contains the segment to detach */
  shm = &shm_region->shm[id];
  p = &shm->ds;

  if (p->shm_nattch)
    p->shm_nattch--;
  if (p->shm_nattch == 0 && shm->flags & SHM_DELSEG) {
    shm_delseg(shm);
  }  


  p->shm_lpid = ipc_pid();
  do_ftime(&dst);
  p->shm_dtime = dst.time;

  shm_leave();
  return 0;
}

/**************************************************************************
 * S E M section
 *************************************************************************/
static char semflg = 0;
static HMTX sem_mutex = 0;
static struct _semcommon *sem_region = 0;

int sem_attach()
{
  int rc;
  struct _semitem *sem;

  if (semflg)
    return sem_region ? -1 : 0;

  rc = sysv_attachglobal("\\SHAREDMEM\\EMXSEM.MEM",
			 "\\SEM32\\EMXSEM.SEM",
			 sizeof(struct _semitem)*SEMMNI + sizeof(short)*2,
			 (PPVOID)&sem_region,
			 &sem_mutex);
  semflg = 1;
  switch (rc) {
  case -1: /* error */
    return -1;
  case 0:  /* opened */
    return 0;
  case 1:  /* created, must be initialized */
    sem_region->seqno = 0;
    sem_region->flags = 0;
    sem_region->usecnt = 0;
    sem = sem_region->sem;
  default:
    return 0;
  }
}

void sem_detach()
{
  sysv_detachglobal(sem_region,sem_mutex);
}

#define sem_enter() DosRequestMutexSem(sem_mutex,-1l)
#define sem_leave() DosReleaseMutexSem(sem_mutex)

int sysv_semctl(struct _semargs* args, int *errnop)
{
  /* not yet supported */
  *errnop = ENOSYS;
  return -1;
}

int sysv_semget(struct _semargs* args, int *errnop)
{
  /* not yet supported */
  *errnop = ENOSYS;
  return -1;
}

int sysv_semop(struct _semargs* args, int *errnop)
{
  /* not yet supported */
  *errnop = ENOSYS;
  return -1;
}

int sysv_semconfig(struct _semargs* args, int *errnop)
{
  /* not yet supported */
  *errnop = ENOSYS;
  return -1;
}

/**************************************************************************
 * M S G section
 *************************************************************************/
static char msgflg = 0;
static HMTX msg_mutex = 0;
static struct _msgcommon *msg_region = 0;

int msg_attach()
{
  int rc;
  struct _msgitem *msg;

  if (msgflg)
    return msg_region ? -1 : 0;

  rc = sysv_attachglobal("\\SHAREDMEM\\EMXMSG.MEM",
			 "\\SEM32\\EMXMSG.SEM",
			 sizeof(struct _msgitem)*MSGMNI + sizeof(short)*2,
			 (PPVOID)&msg_region,
			 &msg_mutex);
  msgflg = 1;
  switch (rc) {
  case -1: /* error */
    return -1;
  case 0:  /* opened */
    return 0;
  case 1:  /* created, must be initialized */
    msg_region->seqno = 0;
    msg_region->flags = 0;
    msg_region->usecnt = 0;
    msg = msg_region->msg;
  default:
    return 0;
  }
}

void msg_detach()
{
  sysv_detachglobal(msg_region,msg_mutex);
}

#define msg_enter() DosRequestMutexSem(msg_mutex,-1l)
#define msg_leave() DosReleaseMutexSem(msg_mutex)

int sysv_msgctl(struct _msgargs* args, int *errnop)
{
  /* not yet supported */
  *errnop = ENOSYS;
  return -1;
}

int sysv_msgget(struct _msgargs* args, int *errnop)
{
  /* not yet supported */
  *errnop = ENOSYS;
  return -1;
}

int sysv_msgrcv(struct _msgargs* args, int *errnop)
{
  /* not yet supported */
  *errnop = ENOSYS;
  return -1;
}

int sysv_msgsnd(struct _msgargs* args, int *errnop)
{
  /* not yet supported */
  *errnop = ENOSYS;
  return -1;
}

/**************************************************************************
 * common functions
 *************************************************************************/

void init_sysvipc(void)
{
  /* actually, don't initialize anything: sysvipc is not generally used,
   * so do the real initialisation on demand only
   */
  shm_region = 0;
  sem_region = 0;
  msg_region = 0;
}

/* this is not really called in init.c */
void term_sysvipc(void)
{
  if (shm_region) shm_detach();
  if (sem_region) sem_detach();
  if (msg_region) msg_detach();
}

/* return -1 if init failed, 0 if shared memory opened, 1 if shared memory
 * newly created
 */
int sysv_attachglobal(const char* memname, const char* mtxname, 
	int memsize, PPVOID memaddr, PHMTX mtxaddr)
{
  APIRET rc,rc2;

  rc = DosGetNamedSharedMem(memaddr,memname,PAG_READ|PAG_WRITE);
  if (rc == ERROR_FILE_NOT_FOUND) {
    /* does not exist, create it */
    rc = DosAllocSharedMem(memaddr,memname,memsize,
                           PAG_COMMIT|PAG_READ|PAG_WRITE);
    if (rc) return -1; /* no such memory */
    else rc = 1;

    /* now get the mutex */
  } else if (rc) /* some other problem */
    return -1;

  /* have the region, now get the mutex */
  rc2 = DosOpenMutexSem(mtxname,mtxaddr);
  if (rc2) {
    /* none found, create it */
    rc2 = DosCreateMutexSem(mtxname,mtxaddr,DC_SEM_SHARED,0);
    if (rc2) {
      /* something went wrong */
      DosFreeMem(*memaddr);
      *memaddr = 0;
      return -1;
    }
  }
  return rc;
}

void sysv_detachglobal(PVOID mem,HMTX mtx)
{
  DosFreeMem(mem);
  DosCloseMutexSem(mtx);
}

int ipc_open(key_t key, int id, int openflg, int fullflg,
		    struct ipc_perm *perm, int* errnop)
{
  if (key == IPC_PRIVATE) {
    if (fullflg || id == -1) {
      *errnop = ENOSPC;
      return -1;
    }
    /* else have an id for a new region */
  } else if (id == -1) { /* key does not exist */
    if (fullflg & IPC_CREAT) {
      if (fullflg) {
        *errnop = ENOSPC;
        return -1;
      }
    } else {
      *errnop = ENOENT;
      return -1;
    }
  } else { /* key exists */
    if (fullflg & IPC_CREAT) {
      if (fullflg && IPC_EXCL) {
        *errnop = EEXIST;
        return -1;
      }
      /* exists, permission okay? */
      if (ipc_access(perm,fullflg & 0777)) {
        *errnop = EACCES;
	return -1;
      }
    }
  }
  return 0;
}

int ipc_access(struct ipc_perm* perm, int flags)
{
  /* this taken from linux, which obviously isn't picky about 
   * exact user/group/other rights either
   */
  int reqst = (flags >> 6) | (flags >> 3) | flags;
  int grant = perm->mode >> 6;
  return (reqst & ~grant & 0007) ? -1 : 0;
}
